// Copyright 1993-2012 Christophe Gottheimer <cgh75015@gmail.com>
//
// MONMON is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 2 as published by the Free
// Software Foundation.  Note that I am not granting permission to redistribute
// or modify nsimII under the terms of any later version of the General Public
// License.
//
// This program is distributed in the hope that it will be useful (or at
// least amusing), but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
// Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program (in the file "COPYING"); if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

static unsigned char monmon[] = {
	0xFD, 0xA8, 0xFD, 0x98, 0xFD, 0x88, 0xFD, 0xC8, 0xFD, 0xAA, 0xFD, 0xC8, 0xB5, 0x01, 0xFD, 0xAE,
	0xF0, 0x0A, 0xFD, 0x48, 0xAA, 0x7B, 0x9F, 0x58, 0x7B, 0x5A, 0xA0, 0x44, 0x6A, 0x09, 0xF5, 0x88,
	0x03, 0x46, 0x84, 0x51, 0x04, 0x51, 0xF4, 0x7B, 0xA8, 0xF6, 0x7B, 0xAC, 0xBE, 0xD0, 0x2B, 0xFD,
	0xE9, 0xF0, 0x0B, 0x00, 0x68, 0x7B, 0xED, 0x76, 0x4F, 0x08, 0x8B, 0x11, 0x6A, 0xA0, 0xBE, 0x2B,
	0xEB, 0x4A, 0x07, 0xBE, 0x2B, 0xD5, 0x4A, 0x04, 0xBE, 0x2B, 0xD5, 0x8E, 0x35, 0x6A, 0xAC, 0x25,
	0x08, 0xBE, 0x2B, 0xEB, 0x25, 0x0A, 0xBE, 0x2B, 0xEB, 0xB5, 0x3A, 0x51, 0xFD, 0x6A, 0x4A, 0x04,
	0xFD, 0xA8, 0xBE, 0x2B, 0xEB, 0xB5, 0x20, 0x51, 0x42, 0x99, 0x09, 0xFD, 0x2A, 0x4A, 0x04, 0xB5,
	0x22, 0x51, 0x65, 0xB9, 0x7F, 0xB7, 0x20, 0x83, 0x02, 0xB5, 0x20, 0x51, 0x42, 0x99, 0x0D, 0xB5,
	0x22, 0x51, 0xB5, 0x80, 0xAE, 0x78, 0x80, 0xBE, 0xE8, 0xCA, 0x5A, 0xB0, 0xBE, 0xE2, 0x4A, 0x18,
	0xF4, 0x7B, 0xAC, 0x5C, 0x0A, 0x89, 0x06, 0x64, 0x64, 0x64, 0x64, 0x9E, 0x74, 0x5C, 0x0B, 0x89,
	0x06, 0x66, 0x66, 0x66, 0x66, 0x9E, 0x7E, 0x5C, 0x08, 0x9B, 0x07, 0x5C, 0x0C, 0x9B, 0x15, 0x5C,
	0x1F, 0x89, 0x0A, 0xA5, 0x76, 0x4F, 0xBD, 0x08, 0xAE, 0x76, 0x4F, 0x9E, 0x91, 0x5C, 0x2E, 0x89,
	0x15, 0xFD, 0xA8, 0xBE, 0x2B, 0xF6, 0xFD, 0x2A, 0xBE, 0xED, 0x95, 0x91, 0xA1, 0x28, 0xBE, 0xED,
	0x95, 0x91, 0xA7, 0x2A, 0x9E, 0xAD, 0x5C, 0x3D, 0x89, 0x0F, 0xFD, 0xA8, 0xBE, 0x2B, 0xF6, 0xFD,
	0x2A, 0xBE, 0xED, 0x95, 0x91, 0xBA, 0x61, 0x9E, 0x08, 0x5C, 0x40, 0x89, 0x08, 0x65, 0x08, 0x25,
	0x0A, 0xFD, 0x6A, 0x9E, 0xCC, 0x5C, 0x1A, 0xCB, 0xFE, 0x8E, 0x58, 0xB5, 0x20, 0x51, 0xBE, 0x2B,
	0xEB, 0x42, 0x99, 0x06, 0x9A, 0xF1, 0xB9, 0x0F, 0xBB, 0x30, 0xB7, 0x3A, 0x81, 0x02, 0xB3, 0x06,
	0x9A, 0x25, 0xBE, 0x2B, 0xDF, 0x51, 0x65, 0xBE, 0x2B, 0xE0, 0x51, 0x9A, 0xBE, 0xD0, 0x2B, 0xB5,
	0x40, 0xAE, 0x78, 0x80, 0xBE, 0xE8, 0xCA, 0xBE, 0xE2, 0x4A, 0xB7, 0x18, 0x9B, 0x12, 0xB7, 0x0E,
	0x8B, 0x08, 0xB7, 0x0D, 0x89, 0x0A, 0x48, 0x7B, 0x4A, 0xB0, 0xFD, 0xE9, 0xF0, 0x0B, 0x00, 0x9A,
	0xB7, 0x30, 0x91, 0x25, 0xB7, 0x47, 0x93, 0x29, 0xB7, 0x3A, 0x81, 0x04, 0xB7, 0x41, 0x91, 0x31,
	0x51, 0x9E, 0x34, 0x5C, 0x3A, 0x99, 0x9C, 0xFD, 0xA8, 0xBE, 0x2B, 0xF6, 0xFD, 0x2A, 0xBE, 0xED,
	0x95, 0x91, 0xA8, 0x18, 0xBE, 0xED, 0x95, 0x91, 0xAE, 0x1A, 0xFD, 0xA8, 0x34, 0x08, 0x0A, 0x65,
	0xFD, 0xCA, 0xA4, 0x96, 0x91, 0x07, 0x89, 0x04, 0x24, 0x16, 0x99, 0x0D, 0xFD, 0x88, 0xBE, 0xD0,
	0x2B, 0xFD, 0x48, 0xFD, 0x6A, 0x64, 0xBE, 0x2B, 0xEB, 0xBE, 0x2B, 0xEB, 0xB5, 0x80, 0xAE, 0x78,
	0x80, 0xBE, 0xE8, 0xCA, 0xBE, 0xE2, 0x4A, 0xFD, 0x2A, 0xFD, 0x2A, 0x9E, 0xE2,
};

static unsigned int reloc16[] = {
	0x2B18, 0x2B1D, 0x2B22, 0x2B2B, 0x2B30, 0x2B3C, 0x2B9D, 0x2BB6,
	0x2BD8, 0x2BEC, 0x2BF1, 0x2C33, 0x2C60, 0x2C63,
};

static unsigned int relocHL[] = {
	0,
};

int main( int argc, char **argv )
{
	unsigned int code = 0x40c5;
	unsigned int tmp, i, offset;
	int fd;
	char name[256];

	if(argc == 2)
		if(1 != sscanf(argv[1], "%x", &code)) {
			fprintf(stderr, "Bad start address\n");
			exit(1);
		}
	printf("Relocating MONMON code to %04X\n", code);
	snprintf(name, sizeof(name), "MONMON-%04X.BIN", code);
	if(0 > (fd = open(name,O_WRONLY|O_CREAT|O_TRUNC,0644))) {
		fprintf(stderr, "Error while opening %s:", name);
		perror("");
		exit(2);
	}
	for(i = 0; i < sizeof(reloc16)/sizeof(unsigned int); i++) {
		tmp = (monmon[reloc16[i]-0x2ADA+1] << 8) | monmon[reloc16[i]-0x2ADA+2];
#ifdef DEBUG
		printf("patch rel16 at %04X = %04X", reloc16[i], tmp);
#endif
		tmp = tmp - 0x2ADA + code;
#ifdef DEBUG
		printf(" -> %04X\n", tmp);
#endif
		monmon[reloc16[i]-0x2ADA+1] = (tmp >> 8) & 0xFF;
		monmon[reloc16[i]-0x2ADA+2] = tmp & 0xFF;
	}
#if 0
	for(i = 0; i < sizeof(relocHL)/sizeof(unsigned int); i++) {
		tmp = (monmon[relocHL[i]-0x2ADA+1] << 8) | monmon[relocHL[i]-0x2ADA-1];
#ifdef DEBUG
		printf("patch relHL at %04X = %04X", reloc16[i], tmp);
#endif
		tmp = tmp - 0x2ADA + code;
#ifdef DEBUG
		printf(" -> %04X\n", tmp);
#endif
		monmon[relocHL[i]-0x2ADA+1] = (tmp >> 8) & 0xFF;
		monmon[relocHL[i]-0x2ADA-1] = tmp & 0xFF;
	}
#endif
	if(sizeof(monmon) != write(fd, monmon, sizeof(monmon))) {
		fprintf(stderr, "Error while writing %s", name);
		perror("");
		close(fd);
		exit(3);
	}
	for(i = 0, tmp = 0; i < sizeof(monmon); i++)
		tmp += monmon[i];
	printf("Written %s, %u bytes, checksum %04X\n", name, i, tmp & 0xFFFF);
	close(fd);
	printf("Call bin2wav %s MONMON-%04X.WAV T:BIN PC:1500 A:0x%04X S:5 N:MONMON-%04X\n",
		name, code, code, code);
	printf("Do a NEW &%04X on the SHARP PC-1500/A or TANDY PC-2\n", code + sizeof(monmon));
	exit(0);
}
